import os
from tools4rdf.network.network import OntologyNetwork


def read_ontology():
    """
    Read in ontologies and perform necessary operations.

    Returns
    -------
    combo: OntologyNetwork, Combined ontology network.
    """
    # read in ontologies
    file_location = os.path.dirname(__file__).split("/")
    file_location = "/".join(file_location[:-1])

    cmso = OntologyNetwork(
        os.path.join(file_location, "data/cmso.owl")
    )  # b7c8f3544d16d0ec9a9aace682da0e6abba1c57f
    pldo = OntologyNetwork(
        os.path.join(file_location, "data/pldo.owl")
    )  # 688ce36a3e6520a6e280bb248789c231d3c36f25
    podo = OntologyNetwork(
        os.path.join(file_location, "data/podo.owl")
    )  # 6a74d511c5b78042e1cb7a6e76e948fa56de598e
    asmo = OntologyNetwork(
        os.path.join(file_location, "data/asmo.owl")
    )  # 688ce36a3e6520a6e280bb248789c231d3c36f25
    ldo = OntologyNetwork(
        os.path.join(file_location, "data/ldo.owl")
    )  # e23fa9930351787e701347878a3e1a0e3924d084

    # combine them
    combo = cmso + pldo + podo + asmo + ldo

    # add namespaces
    combo.add_namespace("prov", "http://www.w3.org/ns/prov#")
    combo.add_namespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
    combo.add_namespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#")

    # add extra terms for quering
    combo.add_term("http://www.w3.org/ns/prov#Entity", "class", delimiter="#")
    combo.add_term("http://www.w3.org/ns/prov#Activity", "class", delimiter="#")
    combo.add_term("http://www.w3.org/ns/prov#SoftwareAgent", "class", delimiter="#")
    combo.add_term(
        "http://www.w3.org/ns/prov#wasDerivedFrom", "object_property", delimiter="#"
    )
    combo.add_term(
        "http://www.w3.org/ns/prov#wasGeneratedBy", "object_property", delimiter="#"
    )
    combo.add_term(
        "http://www.w3.org/ns/prov#wasAssociatedWith", "object_property", delimiter="#"
    )
    combo.add_term(
        "http://www.w3.org/ns/prov#actedOnBehalfOf", "object_property", delimiter="#"
    )
    combo.add_term(
        "http://www.w3.org/2000/01/rdf-schema#label",
        "data_property",
        delimiter="#",
        namespace="rdfs",
        rn=["str"],
    )
    combo.add_term(
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
        "object_property",
        delimiter="#",
        namespace="rdf",
    )

    # add paths

    # General fixes
    combo.add_path(("cmso:CrystalStructure", "cmso:hasAltName", "string"))
    combo.add_path(("cmso:LatticeParameter", "asmo:wasCalculatedBy", "prov:Activity"))

    # interontology paths
    # CMSO -> PODO VACANCY
    combo.add_path(("cmso:Material", "cmso:hasDefect", "pldo:PlanarDefect"))
    combo.add_path(("cmso:Material", "cmso:hasDefect", "podo:Vacancy"))
    combo.add_path(("cmso:AtomicScaleSample", "podo:hasVacancyConcentration", "float"))
    combo.add_path(("cmso:AtomicScaleSample", "podo:hasNumberOfVacancies", "int"))

    # CMSO -> PODO IMPURITY
    combo.add_path(("cmso:Material", "cmso:hasDefect", "podo:SubstitutionalImpurity"))
    combo.add_path(("cmso:Material", "cmso:hasDefect", "podo:InterstitialImpurity"))
    combo.add_path(("podo:InterstitialImpurity", "rdfs:label", "string"))
    combo.add_path(("cmso:AtomicScaleSample", "podo:hasNumberOfImpurityAtoms", "int"))
    combo.add_path(("cmso:AtomicScaleSample", "podo:hasImpurityConcentration", "float"))

    # CMSO -> LDO DISL paths
    combo.add_path(("cmso:Material", "cmso:hasDefect", "ldo:LineDefect"))
    combo.add_path(("cmso:Material", "cmso:hasDefect", "ldo:Dislocation"))

    # CMSO -> PODO DISL paths
    combo.add_path(("cmso:Material", "cmso:hasDefect", "pldo:GrainBoundary"))

    combo.add_path(
        ("cmso:ComputationalSample", "prov:wasDerivedFrom", "cmso:ComputationalSample")
    )
    combo.add_path(("cmso:ComputationalSample", "rdf:type", "prov:Entity"))
    combo.add_path(("cmso:AtomicScaleSample", "prov:wasGeneratedBy", "prov:Activity"))
    combo.add_path(("asmo:EnergyCalculation", "rdf:type", "prov:Activity"))
    combo.add_path(
        ("asmo:EnergyCalculation", "prov:wasAssociatedWith", "prov:SoftwareAgent")
    )
    combo.add_path(
        (
            "cmso:ComputationalSample",
            "prov:wasGeneratedBy",
            "asmo:EnergyCalculation",
        )
    )

    combo.add_path(("cmso:CalculatedProperty", "cmso:hasValue", "float"))
    # combo.add_path(("cmso:CalculatedProperty", "asmo:hasUnit", "string"))
    combo.add_path(("cmso:CalculatedProperty", "rdfs:label", "string"))
    # how to handle units?

    # input parameters for ASMO
    combo.add_path(("asmo:InputParameter", "rdfs:label", "string"))
    # combo.add_path(("cmso:CalculatedProperty", "asmo:hasUnit", "string"))
    combo.add_path(("asmo:InputParameter", "cmso:hasValue", "float"))

    # software agent
    combo.add_path(("prov:SoftwareAgent", "rdfs:label", "string"))
    combo.add_path(("asmo:InteratomicPotential", "cmso:hasReference", "string"))
    combo.add_path(("asmo:InteratomicPotential", "rdfs:label", "string"))

    # now more paths
    combo.add_path(("prov:Activity", "asmo:hasInputParameter", "asmo:InputParameter"))
    combo.add_path(("cmso:CalculatedProperty", "asmo:wasCalculatedBy", "prov:Activity"))
    combo.add_path(
        (
            "cmso:ComputationalSample",
            "cmso:hasCalculatedProperty",
            "cmso:CalculatedProperty",
        )
    )
    combo.add_path(
        (
            "cmso:ComputationalSample",
            "cmso:hasCalculatedProperty",
            "asmo:PhysicalQuantity",
        )
    )
    combo.add_path(
        (
            "cmso:ComputationalSample",
            "cmso:hasCalculatedProperty",
            "asmo:OutputParameter",
        )
    )

    # return
    return combo



"""
https://docs.python.org/3/library/operator.html
"""

from rdflib import URIRef
import numbers
import copy


def _get_namespace_and_name(uri):
    uri_split = uri.split("#")
    if len(uri_split) > 1:
        # possible that delimiter is #
        name = uri_split[-1]
        namespace_split = uri_split[0].split("/")
        namespace = namespace_split[-1]
    else:
        uri_split = uri.split("/")
        if len(uri_split) > 1:
            name = uri_split[-1]
            namespace = uri_split[-2]
        else:
            name = uri_split[-1]
            namespace = ""
    return namespace, name


def _get_namespace_with_prefix(uri):
    uri_split = uri.split("#")
    if len(uri_split) > 1:
        # possible that delimiter is #
        namespace = uri_split[0]
    else:
        uri_split = uri.split("/")
        if len(uri_split) > 1:
            namespace = "/".join(uri_split[:-1])
            if namespace[-1] != "#":
                namespace += "/"
        else:
            namespace = ""
    return namespace


def strip_name(uri, get_what="name", namespace=None):
    if namespace is None:
        namespace, name = _get_namespace_and_name(uri)
    else:
        _, name = _get_namespace_and_name(uri)
    if get_what == "namespace":
        return namespace
    elif get_what == "name":
        return ":".join([namespace, name])
    else:
        raise ValueError("get_what must be either namespace or name")


class OntoTerm:
    def __init__(
        self,
        uri=None,
        namespace=None,
        node_type=None,
        dm=[],
        rn=[],
        data_type=None,
        node_id=None,
        delimiter="/",
        description=None,
        label=None,
        target=None,
        name=None,
    ):

        if uri is None and name is None:
            raise ValueError("Either uri or name must be provided!")
        self.uri = uri
        # type: can be object property, data property, or class
        self.node_type = node_type
        # now we need domain and range
        self.domain = dm
        self.range = rn
        # datatype, that is only need for data properties
        self.data_type = data_type
        # identifier
        self.node_id = node_id
        self.associated_data_node = None
        self.subclasses = []
        self.named_individuals = []
        self.equivalent_classes = []
        self.subproperties = []
        self.delimiter = delimiter
        self._description = None
        self.description = description
        self.label = label
        self._label = None
        self.is_domain_of = []
        self.is_range_of = []
        self._condition = None
        if uri is not None and namespace is None:
            namespace = strip_name(uri, get_what="namespace")
        self.namespace = namespace
        # name of the class
        self._name = None
        self.name = name
        # parents for the class; these are accumulated
        # when using the >> operator
        self._parents = []
        # condition parents are the parents that have conditions
        # these are accumulated when using the & or || operators
        self._condition_parents = []
        self.target = target

    @property
    def URIRef(self):
        return URIRef(self._uri)

    @property
    def uri(self):
        """
        Get the URI of the ontology term.

        Returns
        -------
        str
            The URI of the ontology term.
        """
        return self._uri

    @uri.setter
    def uri(self, val):
        self._uri = val

    @property
    def description(self):
        """
        Get the description of the term.

        Returns
        -------
        str
            The description of the term.
        """
        return self._description

    @description.setter
    def description(self, val):
        if isinstance(val, list):
            if len(val) == 0:
                val = ""
            elif len(val) > 1:
                val = ". ".join(val)
            else:
                val = val[0]
        self._description = val

    @property
    def label(self):
        """
        Get the label of the term.

        Returns
        -------
        str
            The label of the term.
        """
        return self._label

    @label.setter
    def label(self, val):
        if isinstance(val, list):
            if len(val) == 0:
                val = ""
            elif len(val) > 1:
                val = ". ".join(val)
            else:
                val = val[0]
        self._label = val

    @property
    def name(self):
        """
        Get the name of the term.

        Returns
        -------
        str
            The name of the term.
        """
        return self._name

    @name.setter
    def name(self, val):
        if val is None:
            val = strip_name(
                self.uri,
                get_what="name",
                namespace=self.namespace,
            )
        self._name = val

    @property
    def name_without_prefix(self):
        """
        Get the name without the namespace prefix.

        Returns
        -------
        str
            The name of the term without the namespace prefix.
        """
        name = self.name
        name = name.replace("–", "")
        name = name.replace("-", "")
        name = name.split(":")[-1]
        return name

    @property
    def namespace_with_prefix(self):
        """
        Get the namespace of the term with the prefix.

        Returns
        -------
        str
            The namespace of the term with the prefix.
        """
        return _get_namespace_with_prefix(self.uri)

    @property
    def namespace_object(self):
        """
        Get the namespace object for the term.

        Returns
        -------
        object
            The namespace object for the term.

        """
        return self.URIRef

    @property
    def query_name(self):
        """
        Get the name of the term as it appears in a SPARQL query.

        Returns
        -------
        str
            The name of the term in a SPARQL query.

        Notes
        -----
        If the term is a data property, the name will be appended with "value".

        """
        if self.node_type == "data_property":
            return self.name + "value"
        elif self.node_type == "object_property":
            if len(self.range) > 0:
                # this has a domain
                return self.range[0]
        return self.name

    @property
    def variable_name(self):
        """
        Get the name of the term to use as a variable in a SPARQL query.

        Returns
        -------
        str
            The name of the term in a SPARQL query.

        """
        name_list = [x.name_without_prefix for x in self._parents]
        name_list.append(self.name_without_prefix)
        name = "_".join(name_list)

        if self.node_type == "data_property":
            return name + "value"

        return name

    @property
    def query_name_without_prefix(self):
        """
        Get the name of the term as it appears in a SPARQL query without prefix.

        Returns
        -------
        str
            The name of the term in a SPARQL query.

        Notes
        -----
        If the term is a data property, the name will be suffixed with "value".
        """
        if self.node_type == "data_property":
            return self.name_without_prefix + "value"
        return self.name_without_prefix

    def toPython(self):
        return self.uri

    def __repr__(self):
        if self.description is not None:
            return str(self.name + "\n" + self.description)
        else:
            return str(self.name)

    def _clean_datatype(self, r):
        if r == "str":
            return "string"
        return r

    # convenience methods for overload checking
    def _ensure_condition_exists(self):
        if self._condition is None:
            raise ValueError(
                "Individual terms should have condition for this operation!"
            )

    def _is_term(self, val):
        if not isinstance(val, OntoTerm):
            raise TypeError("can only be performed with an OntoTerm!")

    def _is_number(self, val):
        if not isinstance(val, numbers.Number):
            raise TypeError("can only be performed with a number!")

    def _is_data_node(self):
        if not self.node_type == "data_property":
            raise TypeError(
                "This operation can only be performed with a data property!"
            )

    def _create_condition_string(self, condition, val):
        return f'(?{self.variable_name}{condition}"{val}"^^xsd:{self._clean_datatype(self.range[0])})'

    # overloading operators
    def __eq__(self, val):
        """
        =
        """
        # print("eq")
        # print(f'lhs {self} rhs {val}')
        self._is_data_node()
        item = copy.deepcopy(self)
        item._condition = item._create_condition_string("=", val)
        return item

    def __lt__(self, val):
        self._is_number(val)
        self._is_data_node()
        item = copy.deepcopy(self)
        item._condition = item._create_condition_string("<", val)
        return item

    def __le__(self, val):
        self._is_number(val)
        self._is_data_node()
        item = copy.deepcopy(self)
        item._condition = item._create_condition_string("<=", val)
        return item

    def __ne__(self, val):
        # self._is_number(val)
        self._is_data_node()
        item = copy.deepcopy(self)
        item._condition = item._create_condition_string("!=", val)
        return item

    def __ge__(self, val):
        self._is_number(val)
        self._is_data_node()
        item = copy.deepcopy(self)
        item._condition = item._create_condition_string(">=", val)
        return item

    def __gt__(self, val):
        # print("gt")
        # print(f'lhs {self} rhs {val}')
        self._is_number(val)
        self._is_data_node()
        self._condition = self._create_condition_string(">", val)
        return self

    def __and__(self, term):
        self._is_term(term)
        self._is_data_node()
        term._is_data_node()
        self._ensure_condition_exists()
        term._ensure_condition_exists()
        item = copy.deepcopy(self)
        item._condition = "&&".join([item._condition, term._condition])
        item._condition = f"({item._condition})"
        item._condition_parents.append(copy.deepcopy(term))
        # and clean up the inbound term
        if item.name != term.name:
            term.refresh_condition()
        return item

    def and_(self, term):
        self.__and__(term)

    def __or__(self, term):
        self._is_term(term)
        self._is_data_node()
        term._is_data_node()
        self._ensure_condition_exists()
        term._ensure_condition_exists()
        item = copy.deepcopy(self)
        item._condition = "||".join([item._condition, term._condition])
        item._condition = f"({item._condition})"
        item._condition_parents.append(copy.deepcopy(term))
        # and clean up the inbound term
        if item.name != term.name:
            term.refresh_condition()
        return item

    def or_(self, term):
        self.__or__(term)

    def __matmul__(self, term):
        # print("matmul")
        # print(f'lhs {self} rhs {term}')
        item = copy.deepcopy(self)
        item._parents.append(copy.deepcopy(term))
        return item

    def refresh_condition(self):
        self._condition = None
        self._condition_parents = []

    def refresh(self):
        self._condition = None
        self._parents = []
        self._condition_parents = []



"""
These are patches specifically designed for atomrdf.

These may or may not be implemented in the ontology. As it is implemented; it can be removed
from the patches
"""


def patch_terms(iri, rn):
    """
    Remove functions as patching is done
    """
    # do stuff here if you want to patch manually
    return rn



import networkx as nx
import graphviz
import pandas as pd
from rdflib import URIRef, Literal, RDF, OWL

from tools4rdf.network.attrsetter import AttrSetter
from tools4rdf.network.parser import parse_ontology, OntoParser


def _replace_name(name):
    return ".".join(name.split(":"))


def _strip_name(name):
    raw = name.split(":")
    if len(raw) > 1:
        return raw[-1]
    return name


class Network:
    def __init__(self, onto):
        self.terms = AttrSetter()
        self.terms._add_attribute(onto.get_attributes())
        self.g = onto.get_networkx_graph()
        self.namespaces = onto.namespaces
        self.extra_namespaces = onto.extra_namespaces

    def draw(
        self,
        styledict={
            "class": {"shape": "box"},
            "object_property": {"shape": "ellipse"},
            "data_property": {"shape": "ellipse"},
            "literal": {"shape": "parallelogram"},
        },
    ):
        """
        Draw the network graph using graphviz.

        Parameters
        ----------
        styledict : dict, optional
            A dictionary specifying the styles for different node types.
            The keys of the dictionary are the node types, and the values are dictionaries
            specifying the shape for each node type. Defaults to None.

        Returns
        -------
        graphviz.Digraph
            The graph object representing the network graph.

        Example
        -------
        styledict = {
            "class": {"shape": "box"},
            "object_property": {"shape": "ellipse"},
            "data_property": {"shape": "ellipse"},
            "literal": {"shape": "parallelogram"},
        }
        network.draw(styledict)
        """
        dot = graphviz.Digraph()
        node_list = list(self.g.nodes(data="node_type"))
        edge_list = list(self.g.edges)
        for node in node_list:
            name = _replace_name(node[0])
            if node[1] is not None:
                t = node[1]
                dot.node(name, shape=styledict[t]["shape"], fontsize="6")
        for edge in edge_list:
            dot.edge(_replace_name(edge[0]), _replace_name(edge[1]))
        return dot

    def _get_shortest_path(self, source, target):
        # this function will be modified to take OntoTerms direcl as input; and use their names.
        path = nx.shortest_path(
            self.g, source=source.query_name, target=target.query_name
        )
        # replace the start and end with thier corresponding variable names
        path[0] = source.variable_name
        path[-1] = target.variable_name
        return path

    def get_shortest_path(self, source, target, triples=False):
        """
        Compute the shortest path between two nodes in the graph.

        Parameters:
        -----------
        source : node
            The starting node for the path.
        target : node
            The target node for the path.
        triples : bool, optional
            If True, returns the path as a list of triples. Each triple consists
            of three consecutive nodes in the path. If False, returns the path
            as a list of nodes.

        Returns:
        --------
        path : list
            The shortest path between the source and target nodes. If `triples`
            is True, the path is returned as a list of triples. If `triples` is
            False, the path is returned as a list of nodes.

        """
        # this function should also check for stepped queries
        path = []
        if len(target._parents) > 0:
            # this needs a stepped query
            complete_list = [source, *target._parents, target]
            # get path for first two terms
            path = self._get_shortest_path(complete_list[0], complete_list[1])
            for x in range(2, len(complete_list)):
                temp_source = complete_list[x - 1]
                temp_dest = complete_list[x]
                temp_path = self._get_shortest_path(temp_source, temp_dest)
                path.extend(temp_path[1:])
        else:
            path = self._get_shortest_path(source, target)

        if triples:
            triple_list = []
            for x in range(len(path) // 2):
                triple_list.append(path[2 * x : 2 * x + 3])
            return triple_list

        return path

    def create_query(self, source, destinations, enforce_types=True):
        """
        Create a SPARQL query string based on the given source, destinations, condition, and enforce_types.

        Parameters
        ----------
        source : Node
            The source node from which the query starts.
        destinations : list or Node
            The destination node(s) to which the query should reach. If a single
            node is provided, it will be converted to a list.
        enforce_types : bool, optional
            Whether to enforce the types of the source and destination nodes in the query. Defaults to True.

        Returns
        -------
        str
            The generated SPARQL query string.

        """
        # if not list, convert to list
        if not isinstance(destinations, list):
            destinations = [destinations]

        # check if more than one of them have an associated condition -> if so throw error
        no_of_conditions = 0
        for destination in destinations:
            if destination._condition is not None:
                no_of_conditions += 1
        if no_of_conditions > 1:
            raise ValueError("Only one condition is allowed")

        # iterate through the list, if they have condition parents, add them explicitely
        for destination in destinations:
            for parent in destination._condition_parents:
                if parent.variable_name not in [d.variable_name for d in destinations]:
                    destinations.append(parent)

        # all names are now collected, in a list of lists
        # start prefix of query
        query = []
        for key, val in self.namespaces.items():
            query.append(f"PREFIX {key}: <{val}>")
        for key, val in self.extra_namespaces.items():
            query.append(f"PREFIX {key}: <{val}>")

        # construct the select distinct command:
        # add source `variable_name`
        # iterate over destinations, add their `variable_name`
        select_destinations = [
            "?" + destination.variable_name for destination in destinations
        ]
        select_destinations = ["?" + source.variable_name] + select_destinations
        query.append(f'SELECT DISTINCT {" ".join(select_destinations)}')
        query.append("WHERE {")

        # constructing the spaql query path triples, by iterating over destinations
        # for each destination:
        #    - check if it has  parent by looking at `._parents`
        #    - if it has `_parents`, called step path method
        #    - else just get the path
        #    - replace the ends of the path with `variable_name`
        #    - if it deosnt exist in the collection of lines, add the lines
        for count, destination in enumerate(destinations):
            # print(source, destination)
            triplets = self.get_shortest_path(source, destination, triples=True)
            # print(triplets)
            for triple in triplets:
                # print(triple)
                line_text = "    ?%s %s ?%s ." % (
                    triple[0].replace(":", "_"),
                    triple[1],
                    triple[2].replace(":", "_"),
                )
                if line_text not in query:
                    query.append(line_text)

        # we enforce types of the source and destination
        if enforce_types:
            if source.node_type == "class":
                query.append(
                    "    ?%s rdf:type %s ."
                    % (_strip_name(source.variable_name), source.query_name)
                )

            for destination in destinations:
                if destination.node_type == "class":
                    query.append(
                        "    ?%s rdf:type %s ."
                        % (
                            destination.variable_name,
                            destination.query_name,
                        )
                    )
        # - formulate the condition, given by the `FILTER` command:
        #    - extract the filter text from the term
        #    - loop over destinations:
        #        - call `replace(destination.query_name, destination.variable_name)`
        filter_text = ""

        # make filters; get all the unique filters from all the classes in destinations
        for destination in destinations:
            if destination._condition is not None:
                filter_text = destination._condition
                break

        # replace the query_name with variable_name
        if filter_text != "":
            for destination in destinations:
                filter_text = filter_text.replace(
                    destination.query_name, destination.variable_name
                )
            query.append(f"FILTER {filter_text}")
        query.append("}")

        # finished, clean up the terms;
        for destination in destinations:
            destination.refresh()

        return "\n".join(query)

    def query(self, kg, source, destinations, enforce_types=True, return_df=True):
        query_string = self.create_query(
            source, destinations, enforce_types=enforce_types
        )
        res = kg.query(query_string)
        if res is not None:
            if return_df:
                for line in query_string.split("\n"):
                    if "SELECT DISTINCT" in line:
                        break
                labels = [x[1:] for x in line.split()[2:]]
                return pd.DataFrame(res, columns=labels)

        return res


class OntologyNetworkBase(Network):
    """
    Network representation of Onto
    """

    def __init__(self, onto):
        self.onto = onto
        self._terms = None
        self._g = None

    @property
    def terms(self):
        if self._terms is None:
            self._terms = AttrSetter()
            self._terms._add_attribute(self.onto.get_attributes())
        return self._terms

    @property
    def g(self):
        if self._g is None:
            self._g = self.onto.get_networkx_graph()
        return self._g

    def __add__(self, ontonetwork):
        onto = self.onto + ontonetwork.onto
        return OntologyNetworkBase(onto)

    @property
    def attributes(self):
        return self.onto.attributes

    @property
    def namespaces(self):
        return self.onto.namespaces

    @property
    def extra_namespaces(self):
        return self.onto.extra_namespaces

    def __radd__(self, ontonetwork):
        return self.__add__(ontonetwork)

    def add_namespace(self, namespace_name, namespace_iri):
        self.onto.add_namespace(namespace_name, namespace_iri)

    add_namespace.__doc__ = OntoParser.add_namespace.__doc__

    def add_term(
        self,
        uri,
        node_type,
        namespace=None,
        dm=(),
        rn=(),
        data_type=None,
        node_id=None,
        delimiter="/",
    ):
        self.onto.add_term(
            uri=uri,
            node_type=node_type,
            namespace=namespace,
            dm=dm,
            rn=rn,
            data_type=data_type,
            node_id=node_id,
            delimiter=delimiter,
        )
        self._terms = None
        self._g = None

    add_term.__doc__ = OntoParser.add_term.__doc__

    def add_path(self, triple):
        """
        Add a triple as path.

        Note that all attributes of the triple should already exist in the graph.
        The ontology itself is not modified. Only the graph representation of it is.
        The expected use is to bridge between two (or more) different ontologies.

        Parameters
        ----------
        triple : tuple
        A tuple representing the triple to be added. The tuple should contain three elements:
        subject, predicate, and object.

        Raises
        ------
        ValueError
        If the subject or object of the triple is not found in the attributes of the ontology.

        """

        def to_uri(tag, namespaces):
            if ":" in tag:
                prefix, term = tag.split(":")
                return URIRef(namespaces[prefix] + term)
            else:
                return Literal(tag)

        sub, pred, obj = [to_uri(t, self.namespaces) for t in triple]

        if (sub, RDF.type, OWL.Class) not in self.onto.graph:
            raise ValueError(
                f"{sub} not found in {list(self.onto.graph.subjects(RDF.type, OWL.Class))}"
            )

        if (
            isinstance(obj, URIRef)
            and (obj, RDF.type, OWL.Class) not in self.onto.graph
        ):
            raise ValueError(
                f"{obj} not found in {list(self.onto.graph.subjects(RDF.type, OWL.Class))}"
            )

        self.onto.graph.add((sub, pred, obj))
        self._terms = None
        self._g = None


class OntologyNetwork(OntologyNetworkBase):
    """
    Network representation of Onto
    """

    def __init__(self, infile):
        super().__init__(parse_ontology(infile))



import yaml
import os
import numpy as np

doc_fields = ["doc", "url"]


def read_yaml(filename):
    with open(filename, "r") as fin:
        data = yaml.safe_load(fin)
        return data


def _get_doc_from_key(keydict):
    url = keydict["url"] if "url" in keydict.keys() else None

    doc = f"""
    {keydict["doc"]}
    url: {url}
    """
    return doc


class MyList(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AttrSetter:
    """
    Class which enables tab completed, contextual attributes
    """

    def __init__(self):
        self._map_dict = {}
        self.head = None

    def __dir__(self):
        """
        Gives tab completion
        """
        return list(self._map_dict.keys())

    def __repr__(self):
        """
        String representation
        """
        return ", ".join(list(self._map_dict.keys()))

    def __getattr__(self, key):
        """
        Calls attributes from the class when general lookup fails. There are four main types of
        calls that can occur.

        Normal attribute: this function does not get called
        key which is present in `_map_dict`: filter and call it
        key which add ons: for example `position_selected` or `position_not_selected`
        key does not exist: raise `AttributeError`
        """
        if key in self._map_dict.keys():
            if self.head is None:
                self.head = self
            return self._map_dict[key]

        else:
            raise AttributeError(f"Attribute {key} not found")

    def _add_attribute(self, indict, head=None):
        if head is None:
            head = self
        self.head = head

        for key, val in indict.items():
            if isinstance(val, dict):
                if key not in self._map_dict.keys():
                    self._map_dict[key] = AttrSetter()

                self._map_dict[key]._add_attribute(val, head=head)
            else:
                self._map_dict[key] = val



import graphviz
import os
from rdflib import BNode, URIRef, Namespace, Literal
import uuid
import json


def get_title_from_BNode(x):
    return x.toPython()


def get_string_from_URI(x):
    """
    Extract a presentable string from URI.

    Parameters
    ----------
    x : rdflib.term.URIRef
        The URI object to extract the string from.

    Returns
    -------
    tuple
        A tuple containing the presentable string representation of the URI and its type.
        The string representation is the last part of the URI after splitting by '#' or '/'.
        The type can be either "URIRef" or "BNode".
    """
    raw = x.toPython()
    # first try splitting by #
    rawsplit = raw.split("#")
    if len(rawsplit) > 1:
        return rawsplit[-1], "URIRef"

    # try splitting by = for chebi values
    if "CHEBI" in raw:
        rawsplit = raw.split("=")
        rawsplit = rawsplit[-1].split(":")
        if len(rawsplit) > 1:
            return ".".join(rawsplit[-2:]), "URIRef"

    if "sample:" in raw:
        rawsplit = raw.split(":")
        if len(rawsplit) > 1:
            return "_".join(rawsplit), "BNode"

    if "activity:" in raw:
        rawsplit = raw.split(":")
        if len(rawsplit) > 1:
            return "_".join(rawsplit), "BNode"

    if "operation:" in raw:
        rawsplit = raw.split(":")
        if len(rawsplit) > 1:
            return "_".join(rawsplit), "BNode"

    if "property:" in raw:
        rawsplit = raw.split(":")
        if len(rawsplit) > 1:
            return "_".join(rawsplit), "BNode"

    # just a normal url split now
    rawsplit = raw.split("/")
    if len(rawsplit) > 1:
        return ".".join(rawsplit[-2:]), "URIRef"

    # none of the conditions worked, which means it's a hex string
    return raw, "BNode"


def parse_object(x):
    """
    Parse the given object and return its title and type.

    Parameters
    ----------
    x : RDF term
        The RDF term to parse.

    Returns
    -------
    tuple
        A tuple containing the title of the object and its type.

    """
    if isinstance(x, BNode):
        return get_title_from_BNode(x), "BNode"
    elif isinstance(x, URIRef):
        return get_string_from_URI(x)
    elif isinstance(x, Literal):
        return str(x.title()), "Literal"


styledict = {
    "BNode": {"color": "#ffe6ff", "shape": "box", "style": "filled"},
    "URIRef": {"color": "#ffffcc", "shape": "box", "style": "filled"},
    "Literal": {"color": "#e6ffcc", "shape": "ellipse", "style": "filled"},
}


def _switch_box(box):
    if box == "box":
        return "rectangle"
    # remember that only boxes will be used, circles no!


def _fix_id(string1, istype1):
    if istype1 == "Literal":
        id1 = str(uuid.uuid4())
    else:
        id1 = string1
    return id1


def visualize_graph(
    g,
    styledict=styledict,
    rankdir="TB",
    hide_types=False,
    workflow_view=False,
    sample_view=False,
    size=None,
    layout="dot",
):
    """
    Visualizes a graph using Graphviz.

    Parameters
    ----------
    g : dict
        The graph to visualize.
    styledict : dict, optional
        A dictionary containing styles for different types of nodes and edges. Default is `styledict`.
    rankdir : str, optional
        The direction of the graph layout. Default is "TB" (top to bottom).
    hide_types : bool, optional
        Whether to hide nodes with the "type" attribute. Default is False.
    workflow_view : bool, optional
        Whether to enable the workflow view. Default is False.
    sample_view : bool, optional
        Whether to enable the sample view. Default is False.
    size : str, optional
        The size of the graph. Default is None.
    layout : str, optional
        The layout algorithm to use. Default is "dot".

    Returns
    -------
    dot : graphviz.Digraph
        The graph visualization.
    """
    dot = graphviz.Digraph()

    dot.attr(
        rankdir=rankdir,
        style="filled",
        size=size,
        layout=layout,
        overlap="false",
    )

    for k in g:
        string1, istype1 = parse_object(k[0])
        string2, istype2 = parse_object(k[2])
        string3, istype = parse_object(k[1])

        plot = True

        if workflow_view:
            # we collapse sample information
            # if cmso.connector is found, only use it is it is cmso.hasCalculated
            # all sub sample props, indicated by sample_x_jsjsj will be ignored.
            green_list = ["hasCalculatedProperty", "wasCalculatedBy", "hasValue"]
            ssplit = string3.split(".")
            if len(ssplit) == 2:
                if (ssplit[0] == "cmso") and (ssplit[1] not in green_list):
                    plot = False
            if string3 == "subClassOf":
                plot = False
            ssplit = string2.split(".")
            if string3 == "type":
                if (ssplit[0] == "cmso") and (ssplit[1] not in ["CalculatedProperty"]):
                    plot = False
                if (ssplit[0] == "cmso") and (ssplit[1] in ["AtomicScaleSample"]):
                    dot.node(
                        string1,
                        label=string1,
                        shape=styledict[istype1]["shape"],
                        style=styledict[istype1]["style"],
                        color=styledict[istype1]["color"],
                        fontsize=styledict[istype1]["fontsize"],
                        fontname=styledict[istype1]["fontname"],
                    )
                    plot = False
        
        elif sample_view:
            green_list = ['wasDerivedFrom', 'wasGeneratedBy']
            if string3 not in green_list:
                plot = False
            

        if hide_types and (string3 == "type"):
            plot = False

        if not plot:
            continue

        if istype1 == "Literal":
            id1 = str(uuid.uuid4())
        else:
            id1 = string1
        dot.node(
            id1,
            label=string1,
            shape=styledict[istype1]["shape"],
            style=styledict[istype1]["style"],
            color=styledict[istype1]["color"],
            fontsize=styledict[istype1]["fontsize"],
            fontname=styledict[istype1]["fontname"],
        )

        if istype2 == "Literal":
            id2 = str(uuid.uuid4())
        else:
            id2 = string2
        dot.node(
            id2,
            label=string2,
            shape=styledict[istype2]["shape"],
            style=styledict[istype2]["style"],
            color=styledict[istype2]["color"],
            fontsize=styledict[istype2]["fontsize"],
            fontname=styledict[istype2]["fontname"],
        )

        dot.edge(
            id1,
            id2,
            color=styledict["edgecolor"],
            label=string3,
            fontsize=styledict[istype2]["fontsize"],
            fontname=styledict[istype2]["fontname"],
        )

    return dot

def _id(item):
    return str(item).replace(':', '_')

def visualize_provenance(
    prov,
    rankdir="TB",
    size=None,
    layout="dot",
):
    dot = graphviz.Digraph()
    dot.attr(
        rankdir=rankdir,
        style="filled",
        size=size,
        layout=layout,
        overlap="false",
    )
    #add all nodes
    for key in prov.keys():
        nid = _id(key)
        #if "activity" in key:
        dot.node(nid, label=prov[key]['label'], 
                shape='box', 
                color="#C9DAF8", 
                style="filled",
                fontname='Helvetica',
                fontsize='8')
        #else:
        #    dot.node(nid, label=prov[key]['label'], 
        #            shape='parallelogram', 
        #            color="#C9DAF8", 
        #            style="filled",
        #            fontname='Helvetica',
        #            fontsize='8')
    #add all edges
    for key, val in prov.items():
        if 'inputs' in val.keys():
            if val['operation'] == 'input_parameter':
                for subkey, subval in val['inputs'].items():
                    dot.edge(_id(subval), _id(key), label='input_param', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
            if val['operation'] == 'output_parameter':
                for subkey, subval in val['inputs'].items():
                    dot.edge(_id(subval), _id(key), label='output_param', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
            elif val['operation'] == 'sample_for_activity':
                for subkey, subval in val['inputs'].items():
                    dot.edge(_id(subval), _id(key), label='input_sample', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
            elif val['operation'] == 'sample_output':
                for subkey, subval in val['inputs'].items():
                    dot.edge(_id(subval), _id(key), label='output_sample', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
            else:
                operation_id = str(uuid.uuid4())
                operation = dot.node(operation_id, label=val['operation'], 
                                    color="#E6B8AF", 
                                    shape='box', 
                                    style='filled',
                                    fontname='Helvetica',
                                    fontsize='8')
                for subkey, subval in val['inputs'].items():
                    dot.edge(_id(subval), operation_id, label='input', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
                dot.edge(operation_id, _id(key), label='output', 
                        color="#263238",
                        fontname='Helvetica',
                        fontsize='8')
    return dot


from setuptools import setup, find_packages


with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='tools4rdf',
    version='0.2.0',
    author='Abril Azocar Guzman, Sarath Menon',
    author_email='sarath.menon@pyscal.org',
    description='python tool for working with ontologies and data models',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(include=['tools4rdf', 'tools4rdf.*']),
    zip_safe=False,
    download_url = 'https://github.com/ocdo/tools4rdf',
    url = 'https://pyscal.org',
    install_requires=['numpy', 'rdflib', 
    'pyyaml', 'graphviz', 'networkx', 
    'pandas'],
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    include_package_data=True,
)



import os
import networkx as nx

from tools4rdf.network.term import OntoTerm, strip_name
from tools4rdf.network.patch import patch_terms
from rdflib import Graph, RDF, RDFS, OWL, BNode, URIRef


def parse_ontology(infile, format="xml"):
    if not os.path.exists(infile):
        raise FileNotFoundError(f"file {infile} not found!")
    graph = Graph()
    graph.parse(infile, format=format)
    return OntoParser(graph)


class OntoParser:
    def __init__(self, graph):
        self.graph = graph
        self._data_dict = None

    def _initialize(self):
        self._data_dict = {
            "classes": [],
            "attributes": {
                "class": {},
                "object_property": {},
                "data_property": {},
                "data_nodes": {},
            },
            "mappings": {},
            "namespaces": {},
            "extra_namespaces": {},
        }
        self._extract_default_namespaces()
        self.extract_classes()
        self.add_classes_to_attributes()
        self.parse_subclasses()
        self.recursively_add_subclasses()
        self.add_subclasses_to_owlThing()
        self.parse_equivalents()
        self.recursively_add_equivalents()
        self.parse_named_individuals()
        self.extract_object_properties()
        self.extract_data_properties()
        self.recheck_namespaces()

    @property
    def classes(self):
        return self.data_dict["classes"]

    @property
    def mappings(self):
        return self.data_dict["mappings"]

    @property
    def namespaces(self):
        return self.data_dict["namespaces"]

    @property
    def extra_namespaces(self):
        return self.data_dict["extra_namespaces"]

    @property
    def attributes(self):
        return self.data_dict["attributes"]

    @property
    def data_dict(self):
        if self._data_dict is None:
            self._initialize()
        return self._data_dict

    def __add__(self, ontoparser):
        """
        Add method; in principle it should add-
        - classes
        - attributes dict
        """
        graph = self.graph + ontoparser.graph
        return OntoParser(graph)

    def __radd__(self, ontoparser):
        return self.__add__(ontoparser)

    @property
    def base_iri(self):
        base_iri = None
        for s in self.graph.subjects(RDF.type, OWL.Ontology):
            base_iri = str(s)
        return base_iri

    def _extract_default_namespaces(self):
        for prefix, namespace in self.graph.namespaces():
            if len(prefix) > 0 and "default" not in prefix:
                self.namespaces[prefix] = namespace.toPython()

    def recheck_namespaces(self):
        for mainkey in ["class", "object_property", "data_property"]:
            for key, val in self.attributes[mainkey].items():
                namespace = self.attributes[mainkey][key].namespace
                if namespace not in self.namespaces.keys():
                    self.namespaces[namespace] = self.attributes[mainkey][
                        key
                    ].namespace_with_prefix

    def extract_object_properties(self):
        object_properties = list(self.graph.subjects(RDF.type, OWL.ObjectProperty))
        for cls in object_properties:
            term = self.create_term(cls)
            term.domain = self.get_domain(cls)
            term.range = self.get_range(cls)
            term.node_type = "object_property"
            self.attributes["object_property"][term.name] = term

    def extract_data_properties(self):
        data_properties = list(self.graph.subjects(RDF.type, OWL.DatatypeProperty))
        for cls in data_properties:
            term = self.create_term(cls)
            term.domain = self.get_domain(cls)
            rrange = self.get_range(cls)
            rrange = [x.split(":")[-1] for x in rrange]
            rrange = patch_terms(term.uri, rrange)

            term.range = rrange
            term.node_type = "data_property"
            self.attributes["data_property"][term.name] = term

            # now create data nodes
            data_term = OntoTerm(name=term.name + "value", node_type="data_node")
            self.attributes["data_property"][
                term.name
            ].associated_data_node = data_term.name
            self.attributes["data_nodes"][data_term.name] = data_term

    def extract_values(self, subject, predicate):
        for val in self.graph.objects(subject, predicate):
            return val
        return None

    def extract_classes(self):
        for term in self.graph.subjects(RDF.type, OWL.Class):
            if isinstance(term, BNode):
                for relation_type, owl_term in [
                    ("union", OWL.unionOf),
                    ("intersection", OWL.intersectionOf),
                ]:
                    union_term = self.extract_values(term, owl_term)
                    if union_term is not None:
                        unravel_list = self.unravel_relation(union_term)
                        self.mappings[term.toPython()] = {
                            "type": relation_type,
                            "items": [
                                strip_name(item.toPython()) for item in unravel_list
                            ],
                        }
            else:
                self.classes.append(term)
        self.classes.append(URIRef("http://www.w3.org/2002/07/owl#Thing")) 

    def add_classes_to_attributes(self):
        for cls in self.classes:
            term = self.create_term(cls)
            term.node_type = "class"
            self.attributes["class"][term.name] = term

    def get_description(self, cls):
        comment = self.graph.value(
            cls, URIRef("http://purl.obolibrary.org/obo/IAO_0000115")
        )
        if comment is None:
            comment = self.graph.value(
                cls, URIRef("http://www.w3.org/2000/01/rdf-schema#comment")
            )
        if comment is None:
            comment = ""
        return comment

    def lookup_node(self, term):
        if isinstance(term, BNode):
            # lookup needed
            term_name = term.toPython()
            if term_name in self.mappings:
                terms = self.mappings[term_name]["items"]
            else:
                terms = [strip_name(term.toPython())]
        else:
            terms = [strip_name(term.toPython())]
        # so here we map the domain and range wrt to other heirarchies
        additional_terms = []
        # first get subclasses which will share the domain and range
        for term in terms:
            # check if such a thing exists in the class
            if term in self.attributes["class"]:
                # get the subclasses
                additional_terms += self.attributes["class"][term].subclasses
                # get the equivalent classes
                additional_terms += self.attributes["class"][term].equivalent_classes
                # get the named individuals
                additional_terms += self.attributes["class"][term].named_individuals
        # add additiona terms to terms
        terms += additional_terms
        return terms

    def lookup_class(self, term):
        if isinstance(term, BNode):
            term = term.toPython()
        else:
            term = strip_name(term.toPython())
        # print(term)
        if term in self.attributes["class"]:
            return [self.attributes["class"][term].name]
        elif term in self.mappings:
            return self.mappings[term]["items"]
        else:
            return []

    def get_domain(self, cls):
        return self._get_domain_range(cls, RDFS.domain)

    def get_range(self, cls):
        return self._get_domain_range(cls, RDFS.range)

    def _get_domain_range(self, cls, predicate):
        domain = []
        for obj in self.graph.objects(cls, predicate):
            domain_term = self.lookup_node(obj)
            for term in domain_term:
                domain.append(term)
        return domain

    def create_term(self, cls):
        iri = cls.toPython()
        term = OntoTerm(
            uri=iri,
            namespace=self._lookup_namespace(iri),
            description=self.get_description(cls),
            target=cls,
        )
        return term

    def _lookup_namespace(self, uri):
        for key, value in self.namespaces.items():
            if uri.startswith(value):
                return key
        return None

    def unravel_relation(self, term, unravel_list=[]):
        if term == RDF.nil:
            return unravel_list
        first_term = self.graph.value(term, RDF.first)
        if first_term not in unravel_list:
            unravel_list.append(first_term)
        second_term = self.graph.value(term, RDF.rest)
        return self.unravel_relation(second_term, unravel_list)

    def parse_subclasses(self):
        for key, cls in self.attributes["class"].items():
            for obj in self.graph.objects(cls.target, RDFS.subClassOf):
                superclasses = self.lookup_class(obj)
                for superclass in superclasses:
                    self.attributes["class"][superclass].subclasses.append(cls.name)
    
    def recursively_add_subclasses(self):
        for clsname in self.attributes["class"].keys():
            self._recursively_add_subclasses(clsname)

    def _recursively_add_subclasses(self, clsname):
        subclasses_to_add = []
        for subclass in self.attributes["class"][clsname].subclasses:
            for subclass_of_subclass in self.attributes["class"][subclass].subclasses:
                if subclass_of_subclass not in self.attributes["class"][clsname].subclasses:
                    subclasses_to_add.append(subclass_of_subclass)
        if len(subclasses_to_add)==0:
            return
        self.attributes["class"][clsname].subclasses.extend(subclasses_to_add)
        self._recursively_add_subclasses(clsname)

    def add_subclasses_to_owlThing(self):
        for key, cls in self.attributes["class"].items():
            objects = list(self.graph.objects(cls.target, RDFS.subClassOf))
            if len(objects)==0:
                self.attributes["class"]["owl:Thing"].subclasses.append(cls.name)

    def parse_equivalents(self):
        for key, cls in self.attributes["class"].items():
            for equivalent in self.graph.objects(cls.target, OWL.equivalentClass):
                if strip_name(equivalent) in self.attributes["class"]:
                    self.attributes["class"][
                        strip_name(equivalent)
                    ].equivalent_classes.append(cls.name)
                    cls.equivalent_classes.append(strip_name(equivalent))
    
    def recursively_add_equivalents(self):
        for clsname in self.attributes["class"].keys():
            self._recursively_add_equivalents(clsname)

    def _recursively_add_equivalents(self, clsname):
        equivalents_to_add = []
        for equivalent in self.attributes["class"][clsname].equivalent_classes:
            for equivalent_of_equivalent in self.attributes["class"][equivalent].equivalent_classes:
                if equivalent_of_equivalent not in self.attributes["class"][clsname].equivalent_classes:
                    equivalents_to_add.append(equivalent_of_equivalent)
        if len(equivalents_to_add)==0:
            return
        self.attributes["class"][clsname].equivalent_classes.extend(equivalents_to_add)
        self._recursively_add_equivalents(clsname)

    def parse_named_individuals(self):
        named_individuals = list(self.graph.subjects(RDF.type, OWL.NamedIndividual))
        for cls in named_individuals:
            # find parent
            term = self.create_term(cls)
            self.attributes["class"][term.name] = term
            parents = list(self.graph.objects(cls, RDF.type))
            for parent in parents:
                if parent != OWL.NamedIndividual:
                    self.attributes["class"][
                        strip_name(parent.toPython())
                    ].named_individuals.append(term.name)

    def get_attributes(self):
        # add first level - namespaces
        mapdict = {key: {} for key in self.namespaces.keys()}

        # now iterate over all attributes
        for k1 in ["class", "object_property", "data_property"]:
            for k2, val in self.attributes[k1].items():
                mapdict[val.namespace][val.name_without_prefix] = val
        return mapdict

    def get_networkx_graph(self):
        g = nx.DiGraph()
        for key, val in self.attributes["class"].items():
            g.add_node(val.name, node_type="class")

        for property_key in ["object_property", "data_property"]:
            for key, val in self.attributes[property_key].items():
                g.add_node(val.name, node_type=property_key)

                # add edges between them
                for d in val.domain:
                    g.add_edge(d, val.name)

                if property_key == "object_property":
                    for r in val.range:
                        g.add_edge(val.name, r)
                else:
                    g.add_edge(val.name, val.associated_data_node)
        return g

    def add_term(
        self,
        uri,
        node_type,
        namespace=None,
        dm=(),
        rn=(),
        data_type=None,
        node_id=None,
        delimiter="/",
    ):
        """
        Add a node.

        Parameters
        ----------
        uri : str
            The URI of the node.
        node_type : str
            The type of the node.
        namespace : str, optional
            The namespace of the node.
        dm : list, optional
            The domain metadata of the node.
        rn : list, optional
            The range metadata of the node.
        data_type : str, optional
            The data type of the node.
        node_id : str, optional
            The ID of the node.
        delimiter : str, optional
            The delimiter used for parsing the URI.

        Raises
        ------
        ValueError
            If the namespace is not found.

        """
        if node_type == "class":
            self.graph.add((URIRef(uri), RDF.type, OWL.Class))
        elif node_type == "object_property":
            self.graph.add((URIRef(uri), RDF.type, OWL.ObjectProperty))
        elif node_type == "data_property":
            self.graph.add((URIRef(uri), RDF.type, OWL.DatatypeProperty))
        else:
            raise ValueError("Node type not found")
        for r in rn:
            self.graph.add((URIRef(uri), RDFS.range, URIRef(r)))
        for d in dm:
            self.graph.add((URIRef(uri), RDFS.domain, URIRef(d)))
        self._data_dict = None

    def add_namespace(self, namespace_name, namespace_iri):
        """
        Add a new namespace.

        Parameters
        ----------
        namespace_name : str
            The name of the namespace to add.
        namespace_iri : str
            The IRI of the namespace.

        Raises
        ------
        KeyError
            If the namespace already exists.

        """
        if namespace_name not in self.namespaces.keys():
            self.graph.bind(namespace_name, namespace_iri)
        self._data_dict = None



import pytest
from tools4rdf.network.ontology import read_ontology
from rdflib import Graph

def test_network():
    onto = read_ontology()
    kg = Graph()
    kg.parse("tests/triples", format="turtle")
    df = onto.query(kg, onto.terms.cmso.AtomicScaleSample, [onto.terms.cmso.hasSpaceGroupSymbol, onto.terms.cmso.hasNumberOfAtoms==4])
    assert len(df) == 14

def test_owlThing():
    onto = read_ontology()
    query = (onto.create_query(onto.terms.cmso.AtomicScaleSample,onto.terms.cmso.hasAltName@onto.terms.cmso.CrystalStructure))
    assert "CrystalStructure_hasAltNamevalue" in query